#include <string.h>
#include <unistd.h>
#include <iostream>
#include <cstdlib>
#include <cstdio>
#include <sys/time.h>
#include <math.h>
#include <vector>
#include <thread>
#include <sstream>
#include <mutex>
#include <assert.h>
#include <map>

// To compile g++ -g -Wall -pthread -std=c++11 matrix_vector_product.cpp -o matrix_vector_product
// To execute: ./matrix_vector_product nb_threads matrix_size

// To obtain significant speedup (if your implementation is well
// working), use nb_threads = number of cores of your machine and
// matrix_size = 30000 (or even bigger, if your machine has enough
// RAM).

// For simplicity, we assume that the matrix size is a multiple of
// the number of threads.

// We compute the speedup, i.e. the gain of the parallel version with
// respect to the sequential version. If the speedup is higher than
// the number of threads, it means the sequential version performs
// poorly. If the speedup is lower than 1, it means the parallel
// version is slower than the sequential version ! Your objective is
// to propose an approach to get the speedup as close to the number of
// threads as possible.

using namespace std;

// Returns the duration of call
double timevalsub(struct timeval *tv1, const struct timeval *tv2) {
  double res = 0;
  res = tv2->tv_sec - tv1->tv_sec;
  res += (tv2->tv_usec - tv1->tv_usec)*1.0/1000000;
  return res;
}

// Sequential algorithm
void mul_seq(int n, double *a, double *b, double *c){
  for (int i=0; i<n; i++){
    for (int j=0; j<n; j++){
      c[j] += a[i*n+j] * b[i];
    }
  }
}

// Compute the norm of the difference between c1 and c2
double diff_vec(int n, double *c1, double *c2){
  double res = 0.0;
  for (int i=0; i<n; i++){
    res+=(c1[i]-c2[i]) * (c1[i]-c2[i]);
  }
  return sqrt(res);
}

void fct_v0(int id_thread, int slice_size, int n, double *a, double *b, double *c)
{
  // V0: one thread is computed everything, i.e., performs the sequential multiplication
  if (id_thread == 0){
    mul_seq(n, a, b, c);
  }
}

void fct_v1(int id_thread, int slice_size, int n, double *a, double *b, double *c)
{
}

void fct_v2(int id_thread, int slice_size, int n, double *a, double *b, double *c)
{
}

using mult_function = void (*)(int, int, int, double *, double *, double *);
map<string, mult_function> mult_functions_available;
void register_version(mult_function f, string name)
{
  mult_functions_available[name] = f;
}
#define REGISTER_VERSION(f)  \
  do                         \
  {                          \
    register_version(f, #f); \
  } while (0)

int main(int argc, char **argv) {
  REGISTER_VERSION(fct_v0);
  REGISTER_VERSION(fct_v1);
  int nbthreads;
  int n;
  mult_function parallel_f = fct_v0;

  // Reading arguments
  if (argc < 3 || argc > 4){
    cerr << "Usage " << argv[0] << " nb_threads matrix_size [function_for_parallel_computation]"
	 << endl;
    exit(1);    
  }
  nbthreads = atoi(argv[1]);
  n         = atoi(argv[2]);
  cout << "Using " << nbthreads << " threads for a matrix of size " << n
       << endl;

  if (n % nbthreads != 0){
    int new_size = n + (n % nbthreads);
    cout << n << " is not a multiple of the number of threads. Using size=" << new_size << " instead."
         << endl;
    n = new_size;
  }

  if (argc > 3) {
    if (mult_functions_available.find(argv[3]) == mult_functions_available.end()) {
      cout << "No such function '" << argv[3] << "'. Available functions are:" << endl;
      for (auto & f_and_name: mult_functions_available) {
        cout << "   " << f_and_name.first << endl;
      }
      exit(1);
    }
    parallel_f = mult_functions_available[argv[3]];
  }

  cout << "Initialization ... " << flush;

  double *a = (double *) malloc(n*n*sizeof(double));
  double *b = (double *) malloc(n*sizeof(double));
  double *c1 = (double *) malloc(n*sizeof(double));
  double *c2 = (double *) malloc(n*sizeof(double));

  // Initialize a and b to be random
  for (int i=0; i<n*n; i++){
    double r = static_cast <double> (rand()) / (static_cast <double> (RAND_MAX/100));
    a[i] = r;
  }
  for (int i=0; i<n; i++){
    double r = static_cast <double> (rand()) / (static_cast <double> (RAND_MAX/100));
    b[i] = r;
  }
  // Initialize c vectors to 0
  memset(c1, 0, n*sizeof(double));
  memset(c2, 0, n*sizeof(double));

  struct timeval tv1, tv2;
  int err;
  double duree_seq, duree_par;
  cout << "done." << endl;

  // Sequential version
  cout << "Sequential computation ... " << flush;
  err = gettimeofday(&tv1, NULL); assert_perror(err);
  mul_seq(n, a, b, c1);
  err = gettimeofday(&tv2, NULL); assert_perror(err);
  duree_seq = timevalsub(&tv1, &tv2);
  cout << "done." << endl;

  // Parallel version
  cout << "Parallel computation ... " << flush;
  err = gettimeofday(&tv1, NULL); assert_perror(err);

  vector<thread> tab;

  int slice_size = n / nbthreads;
  for (int i=0; i<nbthreads; i++) {
    // Calls the function provided on the command-line (fct_v0 by default).
    // In fct_v0, a single thread performs the full mutiplication
    // TODO: write another function (fct_v1) that follows a basic approach:
    // each thread computes a part of the solution c2.
    // TODO: if the speedup of fct_v1 is not good enough, write another
    // function (fct_v2) to enhance the speedup
    tab.push_back(thread(parallel_f, i, slice_size, n, a, b, c2));
  }
  for (int i=0; i<nbthreads; i++) {
    tab[i].join();
  }

  err = gettimeofday(&tv2, NULL); assert_perror(err);
  cout << "done." << endl;
  duree_par = timevalsub(&tv1, &tv2);

  // If the speedup is higher than the number of threads this is very strange !
  cout << "Sequential: " << duree_seq << "s, Parallel: " << duree_par
       << "s, Speedup: " << duree_seq / duree_par
       << ", Max speedup: " << nbthreads 
       << endl;

  // Check that we have compute the correct thing by comparing the
  // sequential and the parallel result. We are doing many operations,
  // potentially in a different order, so we compute the norm of the
  // difference and do not check that each value is identical.
  double norm_diff = diff_vec(n, c1, c2);
  cout << "Norm of ||c2-c1||: " << norm_diff
       << ", Norm / n should be very small: " << norm_diff / n
       << endl;

  free(a);
  free(b);
  free(c1);
  free(c2);
  
  return 0;
}
